#include "Entity3dCollection.h"

Entity3dCollection::Entity3dCollection(void)
{

	Clear();
}

Entity3dCollection::~Entity3dCollection(void)
{

}

void Entity3dCollection::AddEntity(Entity3d ent, int x, int y, WindowMapping &mapping)
{
	assert(x >= 0);
	assert(y >= 0);
	assert(x < MAX_ENTITIES);
	assert(y < MAX_ENTITIES);
	
	float xpos = x, ypos = y;
	mapping.GetWorldCoordinates(xpos, ypos);
	ent.xpos = xpos;
	ent.ypos = ypos;

	//if(textureIndex >= 0 && textureIndex < TEXTURE_COUNT)
	{
		m_entities[x][y] = ent;

		//connect naturally-connected nodes.

		//first, get links from surrounding nodes to our node.
		ConnectionStruct connections;
		DiscoverConnectedNodes(x, y, connections);

		if(ent.Connectivity.ConnectionsValue)
		{
			//not a boxy structure - corridors.
			//(no need to worry about box connectivity, they never get changed anyway)
			//the strategy is to connect TO boxes.

			if((connections.east) && (ent.Connectivity.Connections.east))
				UnionTwoNodes(x, y, 1, 0);
			if((connections.west) && (ent.Connectivity.Connections.west))
				UnionTwoNodes(x, y, -1, 0);
			if((connections.north) && (ent.Connectivity.Connections.north))
				UnionTwoNodes(x, y, 0, 1);
			if((connections.south) && (ent.Connectivity.Connections.south))
				UnionTwoNodes(x, y, 0, -1);
		}
	}
}

void Entity3dCollection::UnionTwoNodes(int x, int y, int deltax, int deltay)
{
	m_disjointSets.Union(
		m_disjointSets.FindSet(x + MAX_ENTITIES * y),
		m_disjointSets.FindSet(x + deltax + MAX_ENTITIES * (y + deltay)));
}

bool Entity3dCollection::AreDisjoint(int x, int y, int deltax, int deltay)
{
	return 
		m_disjointSets.FindSet(x + MAX_ENTITIES * y) !=
		m_disjointSets.FindSet(x + deltax + MAX_ENTITIES * (y + deltay));
}

//Run this after building the connection.
//Naturally-connected nodes will already exist in the disjoint sets structure (unioned together).
void Entity3dCollection::EnsureAdjacentConnectivity()
{
	for(int x = 0; x < MAX_ENTITIES; x++)
	{
		for(int y = 0; y < MAX_ENTITIES; y++)
		{
			if(m_entities[x][y].rendered_type == RENDERED_NONE)
				continue;

			if(m_entities[x][y].Connectivity.ConnectionsValue == 0)
				continue;

			if(
				(x > 0)
				&& (m_entities[x-1][y].rendered_type != RENDERED_NONE)  //other node exists
				&& (AreDisjoint(x, y, -1, 0))							  //other node disjoint
			  )
			{
				//connect.
				if(m_entities[x-1][y].Connectivity.ConnectionsValue != 0)
					m_entities[x-1][y].Connectivity.Connections.east = 1;
				m_entities[x][y].Connectivity.Connections.west = 1;
				UnionTwoNodes(x, y, -1, 0);
			}

			if(
				(x < MAX_ENTITIES -1)
				&& (m_entities[x+1][y].rendered_type != RENDERED_NONE)  //other node exists
				&& (AreDisjoint(x, y, 1, 0))							  //other node disjoint
			  )
			{
				//connect.
				if(m_entities[x+1][y].Connectivity.ConnectionsValue != 0)
					m_entities[x+1][y].Connectivity.Connections.west = 1;
				m_entities[x][y].Connectivity.Connections.east = 1;
				UnionTwoNodes(x, y, 1, 0);
			}

			if(
				(y > 0)
				&& (m_entities[x][y-1].rendered_type != RENDERED_NONE)  //other node exists
				&& (AreDisjoint(x, y, 0, -1))							  //other node disjoint
			  )
			{
				//connect.
				if(m_entities[x][y-1].Connectivity.ConnectionsValue != 0)
					m_entities[x][y-1].Connectivity.Connections.north = 1;
				m_entities[x][y].Connectivity.Connections.south = 1;
				UnionTwoNodes(x, y, 0, -1);
			}

			if(
				(y < MAX_ENTITIES - 1)
				&& (m_entities[x][y+1].rendered_type != RENDERED_NONE)  //other node exists
				&& (AreDisjoint(x, y, 0, 1))							  //other node disjoint
			  )
			{
				//connect.
				if(m_entities[x][y+1].Connectivity.ConnectionsValue != 0)
					m_entities[x][y+1].Connectivity.Connections.south = 1;
				m_entities[x][y].Connectivity.Connections.north = 1;
				UnionTwoNodes(x, y, 0, +1);
			}
		}
	}
}

void Entity3dCollection::Clear()
{
	m_disjointSets.Reset();

	for(int j = 0; j < MAX_ENTITIES; j++)
	{
		for(int k = 0; k < MAX_ENTITIES; k++)
		{
			m_entities[j][k].rendered_type = RENDERED_NONE;
			m_entities[j][k].Connectivity.ConnectionsValue = 0;
			m_entities[j][k].xpos = 0;
			m_entities[j][k].ypos = 0;
		}
	}
}

void Entity3dCollection::CompileDisplayLists(GLuint firstEntityList, int texturecount)
{
	if(texturecount > TEXTURE_COUNT)
		texturecount = TEXTURE_COUNT;

	for(int tex = 0; tex < texturecount; tex++)
	{
		//build display list
		glNewList(firstEntityList + tex, GL_COMPILE);

		for(int i = 0; i < MAX_ENTITIES; i++)
		{
			for(int j = 0; j < MAX_ENTITIES; j++)
			{
				RenderSingleNode(i, j, tex);
			}
		}

		glEndList();
	}
}

void Entity3dCollection::Render(GLuint *textures, GLuint firstEntityList, int texturecount)
{
	if(texturecount > TEXTURE_COUNT)
		texturecount = TEXTURE_COUNT;

	for(int i = 0; i < texturecount; i++)
	{
		glBindTexture(GL_TEXTURE_2D, textures[i]);
		glCallList(firstEntityList + i);
	}
}


void Entity3dCollection::RenderSingleNode(int x, int y, int texture)
{
	if(m_entities[x][y].rendered_type == RENDERED_NONE)
		return;

	ConnectionStruct connections;
	DiscoverConnectedNodes(x, y, connections);
	ConnectionStruct adjacentTunnels;
	DiscoverConnectedNodes(x, y, adjacentTunnels, true);

	m_entities[x][y].Render(connections, adjacentTunnels, texture);
}

void Entity3dCollection::GetNodeExits(int x, int y, ConnectionStruct &exits)
{
	if(m_entities[x][y].rendered_type == RENDERED_NONE)
	{
		//for now allow traversal in empty space.
		exits.east = exits.north = exits.south = exits.west = 1;
		return;
	}

	ConnectionStruct extconns;
	DiscoverConnectedNodes(x, y, exits);

	//neighbor links only work if we're a "big box" or if we have an outgoing connection as well.
	if(m_entities[x][y].Connectivity.ConnectionsValue != 0)
	{
		exits.north &= m_entities[x][y].Connectivity.Connections.north;
		exits.east &= m_entities[x][y].Connectivity.Connections.east;
		exits.south &= m_entities[x][y].Connectivity.Connections.south;
		exits.west &= m_entities[x][y].Connectivity.Connections.west;
	}
}

void Entity3dCollection::DiscoverConnectedNodes(int x, int y, ConnectionStruct &connections, bool excludeBoxes)
{
	connections.east = connections.north = connections.south = connections.west = 0;

	//some neighbors are not empty.
	if(x > 0)
		connections.west = (m_entities[x - 1][y].rendered_type != RENDERED_NONE) &&
			((m_entities[x - 1][y].Connectivity.Connections.east) ||
			(excludeBoxes ^ (m_entities[x - 1][y].Connectivity.ConnectionsValue == 0)));
	if(x < MAX_ENTITIES - 1)
		connections.east = (m_entities[x + 1][y].rendered_type != RENDERED_NONE) && 
			((m_entities[x + 1][y].Connectivity.Connections.west) ||
			(excludeBoxes ^ (m_entities[x + 1][y].Connectivity.ConnectionsValue == 0)));
	if(y > 0)
		connections.south = (m_entities[x][y - 1].rendered_type != RENDERED_NONE) && 
			((m_entities[x][y - 1].Connectivity.Connections.north) ||
			(excludeBoxes ^ (m_entities[x][y - 1].Connectivity.ConnectionsValue == 0)));
	if(y < MAX_ENTITIES - 1)
		connections.north = (m_entities[x][y + 1].rendered_type != RENDERED_NONE) &&
			((m_entities[x][y + 1].Connectivity.Connections.south) ||
			(excludeBoxes ^ (m_entities[x][y + 1].Connectivity.ConnectionsValue == 0)));
}

bool Entity3dCollection::ConstrainToIndoors(Vector3d &originalPoint, const Vector3d &delta, int gridx, int gridy, int gridx2, int gridy2)
{

	if(m_entities[gridx][gridy].rendered_type == RENDERED_NONE)
	{
		originalPoint = originalPoint + delta;
		return true;
	}

	assert(gridx >= 0);
	assert(gridx < MAX_ENTITIES - 1);
	assert(gridy >= 0);
	assert(gridy < MAX_ENTITIES - 1);

	if((gridx2 >= 0) && (gridx2 < MAX_ENTITIES - 1) && (gridy2 >= 0) && (gridy2 < MAX_ENTITIES - 1))
	{
		if((m_entities[gridx][gridy].rendered_type != RENDERED_NONE) && (m_entities[gridx2][gridy2].rendered_type == RENDERED_NONE))
		{
			LOG("Moving to target would mean leaving the building.");
			return false;
		}
	}
	else
	{
		//this means the next grid square over is off the grid ... shouldn't happen? :D
	}

	ConnectionStruct exits;
	GetNodeExits(gridx, gridy, exits);
	
	ConnectionStruct adjacentTunnels;
	DiscoverConnectedNodes(gridx, gridy, adjacentTunnels, true);

	return m_entities[gridx][gridy].ConstrainToIndoors(originalPoint, delta, exits, adjacentTunnels);
}