#include "GeneratedEntities.h"
#include "World.h"

GLubyte SomeColors[5][3] = {
	{140,	151,	155},
	{202,	132,	0},
	{132,	74,		47},
	{0x44, 0x44, 0},
	{0x44, 0, 0x44}
};

GeneratedEntities::GeneratedEntities(Generator *generator)
{
	m_WindowSizeX = m_WindowSizeY = 8;

	m_EntityCollection.Clear();
	m_Generator = generator;

	m_entityTextures[0] = -1;

	m_firstEntityList = glGenLists(TEXTURE_COUNT);
}

void GeneratedEntities::InitializeTextures()
{
	if(m_entityTextures[0] != -1)
		return;

	for(int i = 0; i < TEXTURE_COUNT; i++)
	{
		//generate textures
		glGenTextures(TEXTURE_COUNT, m_entityTextures);	
	}
}

GeneratedEntities::~GeneratedEntities(void)
{
	Remove();
}

void GeneratedEntities::Remove()
{
	if(m_Generator != NULL)
		delete m_Generator;
	m_Generator = NULL;
}


void GeneratedEntities::Render()
{
	m_EntityCollection.Render(m_entityTextures, m_firstEntityList, m_Generator->GetTextureCount());
}

void GeneratedEntities::Update(bool force)
{
	InitializeTextures();

	if(m_Generator != NULL)
	{
		if(FillCollection(force))
		{
			//set texture
			for(int i = 0; i < TEXTURE_COUNT; i++)
			{
				glBindTexture(GL_TEXTURE_2D, m_entityTextures[i]);
				glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);	// Linear Min Filter
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);	// Linear Mag Filter

				glTexImage2D(GL_TEXTURE_2D, 0, 3, 1, 1, 0, GL_RGB, GL_UNSIGNED_BYTE, SomeColors[i]);

			}

			m_EntityCollection.CompileDisplayLists(m_firstEntityList, m_Generator->GetTextureCount());
		}
	}
}

Generator* GeneratedEntities::GetGenerator()
{
	return m_Generator;
}

int GeneratedEntities::GetId()
{
	if(m_Generator != NULL)
		return m_Generator->id;

	return -1;
}

bool GeneratedEntities::ConstrainToIndoors(Vector3d &originalPoint, const Vector3d &delta)
{
	//current grid
	float gridx, gridy;
	World::theWorld->GetGridContainingPoint(originalPoint, gridx, gridy);
	
	//target grid
	float gridx2, gridy2;
	World::theWorld->GetGridContainingPoint(originalPoint+delta, gridx2, gridy2);

	//transform to physical coordinates
	int x, y;
	m_spaceWindow.GetPhysicalCoordinates(gridx, gridy, x, y);
	
	int x2, y2;
	m_spaceWindow.GetPhysicalCoordinates(gridx2, gridy2, x2, y2);
	return m_EntityCollection.ConstrainToIndoors(originalPoint, delta, x, y, x2, y2);
}

bool GeneratedEntities::FillCollection(bool force)
{
	bool changedWindow = SetSpaceWindow();
	if(!changedWindow && !force) return false;

	m_EntityCollection.Clear();

	//building 3d entity collection
	m_Generator->BuildCollection(&m_EntityCollection, m_spaceWindow);

	return true;
}

bool GeneratedEntities::SetSpaceWindow()
{

	//when the point of interest moves more then 1/window_move_factor, we move the window.
	int window_move_factor = 4;

	//keep divisible by window_move_factor:

	while(m_WindowSizeX % window_move_factor != 0)
		m_WindowSizeX++;

	while(m_WindowSizeY % window_move_factor != 0)
		m_WindowSizeY++;

	//find size of space window
	float x_width = m_WindowSizeX * GRID_SIZE;
	float y_width = m_WindowSizeY * GRID_SIZE;

	float x, y;
	World::theWorld->GetWindowCenter(x, y);

	float x_center, y_center;

	float x_newGrid = GRID_SIZE * x_width*(1/(float)window_move_factor);
	float y_newGrid = GRID_SIZE * y_width*(1/(float)window_move_factor);

	x_center = x_newGrid * floor(x / x_newGrid);
	y_center = y_newGrid * floor(y / y_newGrid);

	WindowMapping tempWindow;
	tempWindow.SetPhysicalWindow(m_WindowSizeX, m_WindowSizeY);
	tempWindow.SetWorldWindow(-y_width/2 + y_center, x_width/2 + x_center, -x_width/2 + x_center, y_width/2 + y_center);

	//no changes - nothing should happen. Window is snapped to grid points.
	if(tempWindow == m_spaceWindow)
		return false;

	LOG("New space window: (%.2f, %.2f), (%.2f, %.2f)", -x_width/2 + x_center, -y_width/2 + y_center, x_width/2 + x_center, y_width/2 + y_center);


	PerformTests(tempWindow);

	//copy constructor
	m_spaceWindow = tempWindow;
	return true;
}

void GeneratedEntities::PerformTests(WindowMapping &mapping)
{
#ifdef _DEBUG
	//performing some "unit testing"
	for(int i = 0; i < mapping.GetPhysicalWidth(); i++)
	{
		for(int j = 0; j < mapping.GetPhysicalHeight(); j++)
		{
			float xpos = i, ypos = j;
			float xtest = xpos, ytest = ypos;
			mapping.GetWorldCoordinates(xtest, ytest);
			int xres, yres;
			mapping.GetPhysicalCoordinates(xtest, ytest, xres, yres);

			if((i != xres) || (j != yres))
				LOG("BAD! (%d, %d) => (%f, %f) => (%d, %d)", i, j, xtest, ytest, xres, yres);
		}
	}
#endif
}

bool GeneratedEntities::SetWindowSize(int newWindowSize)
{
	if(newWindowSize >=2 && m_WindowSizeX != newWindowSize)
	{
		m_WindowSizeX = m_WindowSizeY = newWindowSize;
		return true;
	}

	return false;
}