#include "World.h"
#include "SubGenerator.h"

#include "Timer.h"

World* World::theWorld = 0;

World::World(void)
{
	m_currentLayer = NULL;
	
	for(int i = 0; i < MAX_RENDERABLE; i++)
		m_architecture[i] = NULL;

	m_architectureCount = 0;
	m_visualmode = VISUAL_TOPDOWN;
	m_rendermode = RENDER_SOLID;
	theWorld = this;
}

World::~World(void)
{
	RemoveAllLayers();
}

void World::ResetCameras()
{
	camera.Reset();
	character.Reset();
}

VisualMode World::GetVisualMode()
{
	return m_visualmode;
}

void World::SetVisualMode(VisualMode selected)
{
	m_visualmode = selected;
}

void World::CycleVisualMode()
{
	m_visualmode = (VisualMode) ((m_visualmode + 1) % 2);
}

RenderMode World::GetRenderMode()
{
	return m_rendermode;
}

void World::SetRenderMode(RenderMode selected)
{
	m_rendermode = selected;
}

void World::CreateView()
{
	switch(m_visualmode)
	{
	case VISUAL_TOPDOWN:
		camera.CreateCamera();
		break;
	case VISUAL_CHARACTER:
		character.CreateCamera();
		break;
	}

	switch(m_rendermode)
	{
	case RENDER_SOLID:
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		break;
	case RENDER_WIREFRAME:
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		break;
	}
}

void World::GetGridContainingPoint(const Vector3d &point, float &x, float &y)
{
	GetGridContainingPoint(point.x, point.y, x, y);
}

void World::GetGridContainingPoint(float xpos, float ypos, float &x, float &y)
{
	x = GRID_SIZE * floor((xpos + GRID_SIZE/2) / GRID_SIZE);
	y = GRID_SIZE * floor((ypos + GRID_SIZE/2) / GRID_SIZE);
}

void World::GetCenterPosition(float &x, float &y)
{
	switch(m_visualmode)
	{
	case VISUAL_TOPDOWN:
		{
			x = camera.Lookat.x;
			y = camera.Lookat.y;
			return;
		}
		break;
	case VISUAL_CHARACTER:
	default:
		{
			x = character.Position.x;
			y = character.Position.y;
			return;
		}
		break;
	}
}

void World::GetWindowCenter(float &x, float &y)
{
	float xpos, ypos;
	GetCenterPosition(xpos, ypos);

	GetGridContainingPoint(xpos, ypos, x, y);
}

void World::PanMouse(float dx, float dy)
{
	switch(m_visualmode)
	{
	case VISUAL_TOPDOWN:
		{
			dx *= 10;
			dy *= 10;
			Vector3d thevector = ( camera.Lookat - camera.Position).Normalize();
			if(dx != 0)
			{
				camera.Lookat.x += thevector.y * dx;
				camera.Lookat.y -= thevector.x * dx;

				camera.Position.x += thevector.y * dx;
				camera.Position.y -= thevector.x * dx;
			}
			if(dy != 0)
			{
				camera.Lookat.x -= thevector.x * dy;
				camera.Lookat.y -= thevector.y * dy;

				camera.Position.x -= thevector.x * dy;
				camera.Position.y -= thevector.y * dy;
			}
		}
		break;
	case VISUAL_CHARACTER:
		{
			dx *= 4;
			dy *= 1;

			character.ChangeLook(dx, dy);
		}
		break;
	}

}

void World::ZoomMouse(float dx, float dy)
{
	switch(m_visualmode)
	{
	case VISUAL_TOPDOWN:
		{
			Vector3d outdirection = (camera.Position - camera.Lookat);
			float currentmag = outdirection.Magnitude();
			float newmag = currentmag + dy*10;
			newmag = MAX(newmag, MINIMUM_MAGNITUDE);

			camera.Position = (outdirection.Normalize() * newmag) + camera.Lookat;
		}
		break;
	case VISUAL_CHARACTER:
		{
			//nothing for now, maybe
			break;
		}
	}
}

void World::BuildLighting()
{
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHT1);

	//setting intensity
	float topintensity = 0.5f;
	GLfloat paramsIntensity[] = {topintensity, topintensity, topintensity};
	glLightfv(GL_LIGHT1, GL_DIFFUSE, paramsIntensity);

	//setting position
	GLfloat paramsPosition[] = {camera.Position.x, camera.Position.y, camera.Position.z, 1};
	glLightfv(GL_LIGHT0, GL_POSITION, paramsPosition);

	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial ( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE ) ;

	if(m_visualmode == VISUAL_CHARACTER)
	{
		//add another light
		glEnable(GL_LIGHT2);

		//setting intensity
		float charintensity = 0.3f;
		GLfloat paramsIntensity2[] = {charintensity, charintensity, charintensity};
		glLightfv(GL_LIGHT2, GL_DIFFUSE, paramsIntensity2);

		GLfloat paramsPosition2[] = {character.Position.x, character.Position.y, character.Position.z * 3, 1.0f};
		glLightfv(GL_LIGHT2, GL_POSITION, paramsPosition2);
	}
}

bool World::DoFrame(int timeElapsed, int *keysDown)
{
	bool didChange = false;

	if(m_visualmode == VISUAL_CHARACTER)
	{
		//check keyboard input
		if(keysDown['w'])
		{
			//forward
			character.Go(timeElapsed);
			didChange = true;
		}
		else if(keysDown['s'])
		{
			//backward
			character.Go(-timeElapsed);
			didChange = true;
		}

		//separately
		if(keysDown['a'])
		{
			//left
			character.Strafe(-timeElapsed);
			didChange = true;
		}
		else if(keysDown['d'])
		{
			//right
			character.Strafe(timeElapsed);
			didChange = true;
		}

		if(didChange)
		{
			//make topdown camera follow the character.
			Vector3d diff = character.Position - camera.Lookat;
			camera.Lookat = camera.Lookat + diff;
			camera.Position = camera.Position + diff;
		}
	}

	return didChange;
}

bool World::ConstrainToIndoors(Vector3d &originalPoint, const Vector3d &delta)
{
	//this should loop through all architecture.
	//just use the first one for now.

	bool was_already_inside = true;

	if(m_architectureCount > 0)
	{
		was_already_inside = m_architecture[0]->ConstrainToIndoors(originalPoint, delta);
	}

	return was_already_inside;
}

void World::SetLightingAndCamera()
{
	BuildLighting();
	gluPerspective(60, (GLfloat)m_width / (GLfloat)m_height, 0.01, 100.0);
	CreateView();
}


void World::AddLayer(Generator* newLayer)
{
	if(m_architectureCount < MAX_RENDERABLE - 1 && m_architectureCount >= 0)
	{
		m_architecture[m_architectureCount++] = new GeneratedEntities(newLayer);
		m_currentLayer = newLayer;
	}
}

void World::RemoveLayer(int id)
{
	int index = GetIdIndex(id);

	//does specified index exist in list
	if(index >= 0)
	{
		delete m_architecture[index];
		m_architecture[index] = NULL;

		//move everything back
		for(int i = index; i < (m_architectureCount-1); i++)
		{
			m_architecture[i] = m_architecture[i+1];
			m_architecture[i+1] = NULL;
		}

		m_architectureCount--;


		if(m_architectureCount > 0)
		{
			m_currentLayer = m_architecture[m_architectureCount - 1]->GetGenerator();
		}
		else
		{
			m_currentLayer = NULL;
		}

	}
}

void World::SetSelectedLayer(int selectedID)
{
	if(m_architectureCount <= 0)
	{
		m_currentLayer = NULL;
	}
	else
	{
		int selectedIndex = GetIdIndex(selectedID);
		if(selectedIndex > -1)
			m_currentLayer = m_architecture[selectedIndex]->GetGenerator();
	}
}

int World::GetIdIndex(int id)
{
	for(int i = 0; i < m_architectureCount; i++)
	{
		if(m_architecture[i]->GetId() == id)
			return i;
	}

	return -1;
}

void World::RemoveInputLayer(int selectedIndex)
{
	if(m_currentLayer != NULL)
	{
		m_currentLayer->m_layerVariables[selectedIndex].RemoveRenderable();
	}
}

void World::RemoveAllLayers()
{
	for(int i = m_architectureCount - 1; i >= 0; i--)
		RemoveLayer(m_architecture[i]->GetId());

	m_architectureCount = 0;
}

GeneratorBase* World::GetCurrentLayer()
{
	return m_currentLayer;
}

//expects LayerInfo array at least MAX_RENDERABLE in length.
int World::GetLayers(LayerInfo *infos)
{
	for(int i = 0; i < m_architectureCount; i++)
	{
		Generator* gen = m_architecture[i]->GetGenerator();
		sprintf_s(infos[i].name, "%s %d", RenderableFactory::TypeInformation[gen->m_typeID].type_name, gen->id);
		infos[i].id = gen->id;
	}

	return m_architectureCount;
}

void World::InverseChecked(bool checked)
{
	if(m_currentLayer != NULL)
	{
		if(!IsCurrentLayer3d())
			((SubGenerator*)m_currentLayer)->m_inverse = checked;
	}
}

bool World::IsCurrentLayer3d()
{
	if(!m_currentLayer) return false;

	return RenderableFactory::TypeInformation[m_currentLayer->m_typeID].type_metatype == RENDERABLE_3D;
}

void World::SetInputLayer(int selectedVariable, int typeIndex)
{
	if(m_currentLayer != NULL)
	{
		//ensure that the selected sublayer type is not 3D
		if(RenderableFactory::TypeInformation[typeIndex].type_metatype != RENDERABLE_3D)
		{
			SubGenerator *newLayer = RenderableFactory::CreateLayer2d(typeIndex);

			m_currentLayer->AssignLayerToVariable(selectedVariable, newLayer);
			m_currentLayer = newLayer;
		}
	}
}

void World::SetWindowSize(int newWindowSize, int architectureId)
{
	int index = GetIdIndex(architectureId);
	if(index >= 0 && index < m_architectureCount)
		m_architecture[index]->SetWindowSize(newWindowSize);
}

int World::GetWindowSize(int architectureId)
{
	int index = GetIdIndex(architectureId);
	if(index >= 0 && index < m_architectureCount)
		return m_architecture[index]->m_WindowSizeX;

	return 0;
}

void World::NavigateInto(int selectedVariable)
{
	SubGenerator *sublayer = m_currentLayer->m_layerVariables[selectedVariable].GetRenderable();
	if((m_currentLayer != NULL) && (sublayer != NULL))
	{
		m_currentLayer = sublayer;
	}
}

void World::NavigateUp()
{
	if((m_currentLayer != NULL) && (m_currentLayer->ParentLayer != NULL))
	{
		m_currentLayer = m_currentLayer->ParentLayer;
	}
}

void World::Render()
{
	//3D mode - show everything
	glClearColor (0.35,0.35,0.5,1.0);
	glClear (GL_COLOR_BUFFER_BIT);
	for(int i = 0; i < m_architectureCount; i++)
		m_architecture[i]->Render();

	if(m_visualmode == VISUAL_TOPDOWN)
	{
		character.RenderDirectionIndicator();
	}

#ifdef _DEBUG
	//show a box indicating the current grid square the character is in
	if(m_visualmode == VISUAL_TOPDOWN)
	{
		float gridx, gridy;

		GetGridContainingPoint(character.Position.x, character.Position.y, gridx, gridy);
		int density = 4;
		float radius = 0.05f;
	}
#endif

}

void World::Update(bool force)
{
	for(int i = 0; i < m_architectureCount; i++)
		m_architecture[i]->Update(force);
}

string World::GetLookDirection()
{
	return character.GetLookDirection();
}